/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.core.impl;

import java.sql.*;
import java.util.*;
import org.apache.log4j.*;
import com.silverwrist.util.rcache.ReferencedData;
import com.silverwrist.venice.db.*;
import com.silverwrist.venice.htmlcheck.*;
import com.silverwrist.venice.security.DefaultLevels;
import com.silverwrist.venice.core.*;

class ConferenceUserContextImpl implements ConferenceContext, ConferenceBackend
{
  /*--------------------------------------------------------------------------------
   * Inner class used to cache locally-important data
   *--------------------------------------------------------------------------------
   */

  class ConfCache
  {
    String name;
    String description;
    java.util.Date created;
    int grants_level;

    public ConfCache(String name, String description, java.util.Date created, int grants_level)
    {
      this.name = name;
      this.description = description;
      this.created = created;
      this.grants_level = grants_level;

    } // end constructor

    public String getName()
    {
      return name;

    } // end getName

    public String getDescription()
    {
      return description;

    } // end getDescription

    public java.util.Date getCreated()
    {
      return created;

    } // end getCreated

    public int getGrantsLevel()
    {
      return grants_level;

    } // end getGrantsLevel

  } // end class ConfCache

  /*--------------------------------------------------------------------------------
   * Internal class used in the implementation of fixSeen()
   *--------------------------------------------------------------------------------
   */

  static class FixSeenHelper
  {
    private int topicid;
    private int top_message;
    private boolean do_insert;

    public FixSeenHelper(int topicid, int top_message, boolean do_insert)
    {
      this.topicid = topicid;
      this.top_message = top_message;
      this.do_insert = do_insert;

    } // end constructor

    public void doFix(Statement stmt, int uid, java.util.Date date) throws SQLException
    {
      StringBuffer sql = new StringBuffer();
      if (do_insert)
      { // construct an SQL INSERT statement
	sql.append("INSERT INTO topicsettings (topicid, uid, last_message, last_read) VALUES (");
	sql.append(topicid).append(", ").append(uid).append(", ").append(top_message).append(", '");
	sql.append(SQLUtil.encodeDate(date)).append("');");

      } // end if
      else
      { // construct an SQL UPDATE statement
	sql.append("UPDATE topicsettings SET last_message = ").append(top_message).append(", last_read = '");
	sql.append(SQLUtil.encodeDate(date)).append("' WHERE topicid = ").append(topicid);
	sql.append(" AND uid = ").append(uid).append(';');

      } // end else

      // now execute the update
      stmt.executeUpdate(sql.toString());

    } // end doFix

  } // end class FixSeenHelper

  /*--------------------------------------------------------------------------------
   * Static class used to return hot list entries
   *--------------------------------------------------------------------------------
   */

  static class MyHotlist implements ConferenceHotlistEntry
  {
    private ConferenceContext conf;
    private int seq;

    MyHotlist(ConferenceContext conf, int seq)
    {
      this.conf = conf;
      this.seq = seq;

    } // end constructor

    public ConferenceContext getConference()
    {
      return conf;

    } // end getConference

    public int getSequence()
    {
      return seq;

    } // end getSequence

  } // end class MyHotlist

  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(ConferenceUserContextImpl.class.getName());

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private EngineBackend engine;       // pointer to the engine back end
  private SIGBackend sig;             // pointer to the SIG back end
  private DataPool datapool;          // pointer to the data pool
  private int confid;                 // the conference ID
  private int level;                  // access level I have within this conference
  private ConfCache cache;            // cached internal data for the conference
  private ConferenceSIGContext confdata = null;  // the internal subobject
  private boolean settings_loaded = false;  // did we load the settings from the database?
  private String pseud;               // default pseud to use in this conference
  private java.util.Date last_read;   // last date we read this conference
  private java.util.Date last_post;   // last date we posted to this conference
  private boolean deleted = false;    // has conference been deleted?

  /*--------------------------------------------------------------------------------
   * Constructors
   *--------------------------------------------------------------------------------
   */

  protected ConferenceUserContextImpl(EngineBackend engine, SIGBackend sig, DataPool datapool, int confid,
				      String name, String description, java.util.Date created, int sig_level,
				      int user_level, Connection conn) throws DataException
  {
    this.engine = engine;
    this.sig = sig;
    this.datapool = datapool;
    this.confid = confid;
    this.cache = new ConfCache(name,description,created,sig_level);
    recalcLevel(user_level);
    tryLoadSettings(conn);

  } // end constructor

  ConferenceUserContextImpl(EngineBackend engine, SIGBackend sig, DataPool datapool,
			    ConferenceSIGContext cdata) throws DataException
  {
    this.engine = engine;
    this.sig = sig;
    this.datapool = datapool;
    this.confid = cdata.getConfID();
    this.cache = null;
    cdata.rd_addRef();
    this.confdata = cdata;
    recalcLevel(DefaultLevels.hostConference());
    this.pseud = sig.userDefaultPseud();
    this.last_read = null;
    this.last_post = null;

  } // end constructor

  /*--------------------------------------------------------------------------------
   * finalize() function
   *--------------------------------------------------------------------------------
   */

  protected void finalize()
  {
    if (confdata!=null)
      confdata.rd_release();
    engine = null;
    sig = null;
    datapool = null;
    cache = null;
    confdata = null;

  } // end finalize

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private ConferenceSIGContext getConferenceData() throws DataException
  {
    if (confdata==null)
    { // attempt to load the ConferenceSIGContext
      if (deleted)
	throw new DataException("This conference has been deleted.");
      confdata = sig.getConferenceDataObject(confid);
      if (confdata!=null)
	sig.saveMRU("conf",confdata);

      // clear cache when we get the real confdata
      cache = null;

    } // end if

    return confdata;

  } // end getConferenceData

  private ConferenceSIGContext getConferenceDataNE()
  {
    if (confdata==null)
    { // we need to load the ConferenceSIGContext...
      if (deleted)
	return null;  // no return on deletion

      try
      { // attempt to load the ConferenceSIGContext
	confdata = sig.getConferenceDataObject(confid);
	if (confdata!=null)
	  sig.saveMRU("conf",confdata);

      } // end try
      catch (DataException e)
      { // there was a data failure loading the conference
	logger.error("** Got DataException in getConferenceDataNE: " + e.getMessage(),e);
	return null;

      } // end catch

      // clear cache when we get the real confdata
      cache = null;

    } // end if

    return confdata;

  } // end getConferenceDataNE

  private void recalcLevel(int new_level)
  {
    level = sig.realSIGLevel();

    int sig_level = -1;
    if (cache!=null)
      sig_level = cache.getGrantsLevel();
    else
    { // look at the ConferenceSIGContext to get the SIG level
      ConferenceSIGContext c = getConferenceDataNE();
      if (c!=null)
	sig_level = c.getSIGGrantedLevel();

    } // end else

    if (sig_level>level)
      level = sig_level;
    if (new_level>level)
      level = new_level;

  } // end recalcLevel

  private void tryLoadSettings(Connection conn) throws DataException
  {
    if (deleted)
      return;  // this is a no-op

    try
    { // look up the conference settings for this user
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT * FROM confsettings WHERE confid = ");
      sql.append(confid).append(" AND uid = ").append(sig.realUID()).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (rs.next())
      { // load in the settings
	settings_loaded = true;
	pseud = rs.getString("default_pseud");
	last_read = SQLUtil.getFullDateTime(rs,"last_read");
	last_post = SQLUtil.getFullDateTime(rs,"last_post");

      } // end if
      else
      { // just default everything
	settings_loaded = false;
	pseud = sig.userDefaultPseud();
	last_read = null;
	last_post = null;

      } // end else

    } // end try
    catch (SQLException e)
    { // handle a SQL error as usual
      logger.error("Error loading conference settings: " + e.getMessage(),e);
      throw new DataException("Unable to load conference settings: " + e.getMessage(),e);

    } // end catch

  } // end tryLoadSettings

  /*--------------------------------------------------------------------------------
   * Implementations from interface ConferenceContext
   *--------------------------------------------------------------------------------
   */

  public int getConfID()
  {
    return confid;

  } // end getConfID

  public String getName()
  {
    if (cache!=null)
      return cache.getName();
    ConferenceSIGContext c = getConferenceDataNE();
    if (c!=null)
      return c.getName();
    else
      return null;

  } // end getName
 
  public String getDescription()
  {
    if (cache!=null)
      return cache.getDescription();
    ConferenceSIGContext c = getConferenceDataNE();
    if (c!=null)
      return c.getDescription();
    else
      return null;

  } // end getDescription

  public java.util.Date getCreationDate()
  {
    if (cache!=null)
      return cache.getCreated();
    ConferenceSIGContext c = getConferenceDataNE();
    if (c!=null)
      return c.getCreationDate();
    else
      return null;

  } // end getCreationDate

  public java.util.Date getLastUpdateDate()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c!=null)
      return c.getLastUpdateDate();
    else
      return null;

  } // end getLastUpdateDate

  public List getAliases() throws DataException
  {
    return getConferenceData().getAliases();

  } // end getAliases

  public List getHosts() throws DataException
  {
    return getConferenceData().getHosts();

  } // end getHosts

  public boolean canReadConference()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canReadConference(level);

  } // end canReadConference

  public boolean canPostToConference()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canPostToConference(level);

  } // end canPostToConference

  public boolean canCreateTopic()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canCreateTopic(level);

  } // end canCreateTopic

  public boolean canChangeConference()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canChangeConference(level);

  } // end canChangeConference

  public int getReadLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getReadLevel();

  } // end getReadLevel

  public int getPostLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getPostLevel();

  } // end getPostLevel

  public int getCreateLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getCreateLevel();

  } // end getCreateLevel

  public int getHideLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getHideLevel();

  } // end getHideLevel

  public int getNukeLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getNukeLevel();

  } // end getNukeLevel

  public int getChangeLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getChangeLevel();

  } // end getChangeLevel

  public int getDeleteLevel() throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return getConferenceData().getDeleteLevel();

  } // end getDeleteLevel

  public void setSecurityLevels(int read, int post, int create, int hide, int nuke, int change, int delete)
      throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to change security info");
      throw new AccessError("You are not permitted to change security information for this conference.");

    } // end if

    // "Correct" the security levels specified.
    if (post>create)
    { // Post level should be no greater than Create
      if (logger.isDebugEnabled())
	logger.debug("Post level (" + post + ") was greater than create level (" + create + "), correcting");
      post = create;

    } // end if

    if (read>post)
    { // Read level should be no greater than Post
      if (logger.isDebugEnabled())
	logger.debug("Read level (" + read + ") was greater than post level (" + post + "), correcting");
      read = post;

    } // end if

    if (change>delete)
    { // Change level should be no greater than Delete
      if (logger.isDebugEnabled())
	logger.debug("Change level (" + change + ") was greater than delete level (" + delete
		     + "), correcting");
      change = delete;

    } // end if

    if (nuke>delete)
    { // Nuke level should be no greater than Delete
      if (logger.isDebugEnabled())
	logger.debug("Nuke level (" + nuke + ") was greater than delete level (" + delete + "), correcting");
      nuke = delete;

    } // end if

    if (hide>nuke)
    { // Hide level should be no greater than Nuke
      if (logger.isDebugEnabled())
	logger.debug("Hide level (" + hide + ") was greater than nuke level (" + nuke + "), correcting");
      hide = nuke;

    } // end if

    // Now call through to set the security levels.
    getConferenceData().setSecurityLevels(sig,read,post,create,hide,nuke,change,delete);

  } // end setSecurityLevels

  public void setName(String val) throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to change name");
      throw new AccessError("You are not permitted to change the name of this conference.");

    } // end if

    // Call through to set the name.
    getConferenceData().setName(sig,val);

  } // end setName

  public void setDescription(String val) throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to change description");
      throw new AccessError("You are not permitted to change the description of this conference.");

    } // end if

    // Call through to set the name.
    getConferenceData().setDescription(val);

  } // end setDescription

  public void addAlias(String alias) throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to change aliases");
      throw new AccessError("You are not permitted to change the aliases of this conference.");

    } // end if

    // Call through to add the alias.
    getConferenceData().addAlias(sig,alias);

  } // end addAlias

  public void removeAlias(String alias) throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to change aliases");
      throw new AccessError("You are not permitted to change the aliases of this conference.");

    } // end if

    // Call through to remove the alias.
    getConferenceData().removeAlias(sig,alias);

  } // end removeAlias

  public void setMembership(int uid, int grant_level) throws DataException, AccessError
  {
    if (!(getConferenceData().canChangeConference(level)))
    { // this user can't modify the conference security
      logger.error("user not permitted to change membership");
      throw new AccessError("You are not permitted to change the membership of this conference.");

    } // end if

    // call down to set the level
    getConferenceData().setMembership(sig,uid,grant_level);

    // If we adjusted our OWN level, reflect that change.
    if (uid==sig.realUID())
      recalcLevel(grant_level);

  } // end setMembership

  public void addMember(int uid, boolean as_host) throws DataException, AccessError
  {
    setMembership(uid,(as_host ? DefaultLevels.hostConference() : DefaultLevels.memberConference()));

  } // end addMember

  public void removeMember(int uid) throws DataException, AccessError
  {
    setMembership(uid,-1);

  } // end removeMember

  public int getSIGGrantedLevel() throws DataException, AccessError
  {
    ConferenceSIGContext c = getConferenceData();
    if (!(sig.userCanCreateSubobjects()))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return c.getSIGGrantedLevel();

  } // end getSIGGrantedLevel

  public void setSIGGrantedLevel(int new_level) throws DataException, AccessError
  {
    ConferenceSIGContext c = getConferenceData();
    if (!(sig.userCanCreateSubobjects()))
    { // this user can't modify the conference security
      logger.error("user not permitted to change security info");
      throw new AccessError("You are not permitted to change security information for this conference.");

    } // end if

    c.setSIGGrantedLevel(sig,new_level);

  } // end setSIGGrantedLevel

  public short getSequence() throws DataException, AccessError
  {
    ConferenceSIGContext c = getConferenceData();
    if (!(sig.userCanCreateSubobjects()))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve layout info");
      throw new AccessError("You are not permitted to retrieve layout information for this conference.");

    } // end if

    return c.getSequence();

  } // end getSequence

  public void setSequence(short seq) throws DataException, AccessError
  {
    ConferenceSIGContext c = getConferenceData();
    if (!(sig.userCanCreateSubobjects()))
    { // this user can't modify the conference security
      logger.error("user not permitted to set layout info");
      throw new AccessError("You are not permitted to set layout information for this conference.");

    } // end if

    c.setSequence(seq);

  } // end setSequence

  public boolean getHideList() throws DataException, AccessError
  {
    ConferenceSIGContext c = getConferenceData();
    if (!(sig.userCanCreateSubobjects()))
    { // this user can't modify the conference security
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this conference.");

    } // end if

    return c.getHideList();

  } // end getHideList

  public void setHideList(boolean flag) throws DataException, AccessError
  {
    ConferenceSIGContext c = getConferenceData();
    if (!(sig.userCanCreateSubobjects()))
    { // this user can't modify the conference security
      logger.error("user not permitted to change security info");
      throw new AccessError("You are not permitted to change security information for this conference.");

    } // end if

    c.setHideList(sig,flag);

  } // end setHideList

  public String getDefaultPseud()
  {
    return pseud;

  } // end getDefaultPseud

  public void setDefaultPseud(String val) throws DataException
  {
    if (sig.userIsAnonymous())
      return;  // anonymous user can't change pseud
    if (deleted)
      throw new DataException("This conference has been deleted.");

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer();

      if (settings_loaded)
      { // generate an update statement
	sql.append("UPDATE confsettings SET default_pseud = '").append(SQLUtil.encodeString(val));
	sql.append("' WHERE confid = ").append(confid).append(" AND uid = ").append(sig.realUID()).append(';');

      } // end if
      else
      { // need to insert a confsettings row
	sql.append("INSERT INTO confsettings (confid, uid, default_pseud) VALUES (").append(confid);
	sql.append(", ").append(sig.realUID()).append(", '").append(SQLUtil.encodeString(val)).append("');");

      } // end else

      // execute the statement
      stmt.executeUpdate(sql.toString());

      // save off the values to our local fields
      pseud = val;
      settings_loaded = true;

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error setting default pseud: " + e.getMessage(),e);
      throw new DataException("unable to set default pseud: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setDefaultPseud

  public boolean anyUnread()
  {
    if (deleted || sig.userIsAnonymous())
      return false;
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();

      // Build a query.  The idea here is that we want to see the topic IDs of all topics within the
      // conference which (a) are not hidden and (b) have one or more messages unread by this user.
      // The need for the user's settings for both (a) and (b) necessitates the use of the LEFT JOIN
      // to topicsettings, which in turn necessitates the IFNULL guards on references to topicsettings
      // columns in the WHERE clause (as there's only a topicsettings row if the user has read anything
      // in this topic or otherwise set it).
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT t.topicid FROM topics t LEFT JOIN topicsettings s "
					  + "ON t.topicid = s.topicid AND s.uid = ");
      sql.append(sig.realUID()).append(" WHERE t.confid = ").append(confid);
      sql.append(" AND IFNULL(s.hidden,0) = 0 AND (t.top_message - IFNULL(s.last_message,-1)) > 0 LIMIT 1;");
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // run the query
      ResultSet rs = stmt.executeQuery(sql.toString());
      return rs.next();  // will return 0 or 1 rows

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error getting unread flag: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return false;  // default return

  } // end anyUnread

  public List getTopicList(int get_option, int sort_option) throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to get topics");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if
    
    // palm it all off on the big static function
    return TopicUserContextImpl.getTopicList(engine,this,datapool,get_option,sort_option);

  } // end getTopicList

  public TopicContext getTopic(short number) throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to change membership");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    // call down to the TopicUserContextImpl implementation
    return TopicUserContextImpl.getTopicByNumber(engine,this,datapool,number);

  } // end getTopic

  public TopicContext addTopic(String title, String zp_pseud, String zp_text)
      throws DataException, AccessError
  {
    if (!(getConferenceData().canCreateTopic(level)))
    { // not allowed to create a topic - bail out
      logger.error("user not permitted to create new topic");
      throw new AccessError("You are not permitted to create a topic in this conference.");

    } // end if

    // preprocess the pseud and title arguments through HTML checkers
    HTMLChecker title_ch = engine.createCheckerObject(engine.HTMLC_POST_PSEUD);
    HTMLChecker zp_pseud_ch = engine.createCheckerObject(engine.HTMLC_POST_PSEUD);

    try
    { // run arguments through the HTML checker
      title_ch.append(title);
      title_ch.finish();
      zp_pseud_ch.append(zp_pseud);
      zp_pseud_ch.finish();

    } // end try
    catch (AlreadyFinishedException e)
    { // this isn't right...
      throw new InternalStateError("HTMLChecker erroneously throwing AlreadyFinishedException",e);

    } // end catch

    // Create the actual topic in the database.
    String real_title;
    ReturnTopicInfo new_topic_inf;
    try
    { // call down to create the new topic!
      real_title = title_ch.getValue();
      new_topic_inf = getConferenceData().createNewTopic(sig,real_title,zp_pseud_ch.getValue(),zp_text);

    } // end try
    catch (NotYetFinishedException e)
    { // this isn't right either!
      throw new InternalStateError("HTMLChecker erroneously throwing NotYetFinishedException",e);

    } // end catch

    // now we need to reset our last post date
    Connection conn = null;
    try
    { // get a connection
      conn = datapool.getConnection();

      // create a new record in topicsettings (we WERE the first to post in the topic after all!)
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("INSERT INTO topicsettings (topicid, uid, last_post) VALUES (");
      sql.append(new_topic_inf.getTopicID()).append(", ").append(sig.realUID()).append(", '");
      sql.append(SQLUtil.encodeDate(new_topic_inf.getCreateDate())).append("');");
      stmt.executeUpdate(sql.toString());

      // update the conference last-post information
      touchPost(conn,new_topic_inf.getCreateDate());

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error updating user information: " + e.getMessage(),e);
      throw new DataException("unable to update user information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    // create the topic context to return to the user
    return new TopicUserContextImpl(engine,this,datapool,new_topic_inf,real_title);

  } // end addTopic

  public TopicMessageContext getMessageByPostID(long postid) throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to get messages");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    // call down to the static function level
    return TopicMessageUserContextImpl.getMessage(engine,this,datapool,postid);

  } // end getMessageByPostID

  public HTMLChecker getNewTopicPreviewChecker()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    short new_topic;
    if (c==null)
      new_topic = 0;
    else
      new_topic = (short)(c.getTopTopic() + 1);
    HTMLChecker rc = engine.createCheckerObject(engine.HTMLC_PREVIEW_BODY);
    rc.setContextValue("PostLinkDecoderContext",createDecoderContext(new_topic));
    return rc;
    
  } // end getNewTopicPreviewChecker

  public void fixSeen() throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to set unread messages");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    if (sig.userIsAnonymous())
      return;  // anonymous user can't fixseen

    Connection conn = null;

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // lock the tables that we need
      stmt.executeUpdate("LOCK TABLES confsettings WRITE, topicsettings WRITE, topics READ;");

      try
      { // get the existing topic list and its top messages list
	StringBuffer sql =
          new StringBuffer("SELECT topics.topicid, topics.top_message, ISNULL(topicsettings.last_message) "
			   + "FROM topics LEFT JOIN topicsettings ON topics.topicid = topicsettings.topicid "
			   + "AND topicsettings.uid = ");
	sql.append(sig.realUID()).append(" WHERE topics.confid = ").append(confid).append(';');
	ResultSet rs = stmt.executeQuery(sql.toString());

	// use the results to build up a list of FixSeenHelpers
	ArrayList tmp = new ArrayList();
	while (rs.next())
	  tmp.add(new FixSeenHelper(rs.getInt(1),rs.getInt(2),rs.getBoolean(3)));

	// now iterate over the list and call doFix on each one
	Iterator it = tmp.iterator();
	java.util.Date now = new java.util.Date();
	while (it.hasNext())
	{ // just hit each one in turn
	  FixSeenHelper fsh = (FixSeenHelper)(it.next());
	  fsh.doFix(stmt,sig.realUID(),now);

	} // end while

	// update our last-read indicator, too
	touchRead(conn);

      } // end try
      finally
      { // make sure we unlock everything before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error updating user information: " + e.getMessage(),e);
      throw new DataException("unable to update user information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end fixSeen

  public List getActivePosters(int skip, int limit) throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to get posters list");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    Connection conn = null;
    ArrayList rc = new ArrayList();

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create the SQL statement to retrieve all posters
      StringBuffer sql =
          new StringBuffer("SELECT s.uid, u.username, s.last_read, s.last_post FROM confsettings s, "
			   + "users u WHERE u.uid = s.uid AND s.confid = ");
      sql.append(confid).append(" AND u.is_anon = 0 AND ISNULL(s.last_post) = 0 ORDER BY s.last_post DESC");
      if ((skip>=0) && (limit>0))
	sql.append(" LIMIT ").append(skip).append(", ").append(limit);
      sql.append(';');

      // execute the statement
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // return all the records as ActiveUser data elements
	ActiveUser usr = new ActiveUserImpl(rs.getInt(1),rs.getString(2),SQLUtil.getFullDateTime(rs,3),
					    SQLUtil.getFullDateTime(rs,4));
	rc.add(usr);

      } // end while

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error getting active poster list: " + e.getMessage(),e);
      throw new DataException("unable to get active poster listing: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return Collections.unmodifiableList(rc);

  } // end getActivePosters

  public List getActivePosters(int limit) throws DataException, AccessError
  {
    return getActivePosters(0,limit);

  } // end getActivePosters

  public List getActivePosters() throws DataException, AccessError
  {
    return getActivePosters(-1,-1);

  } // end getActivePosters

  public List getActiveReaders(int skip, int limit) throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to get readers list");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    Connection conn = null;
    ArrayList rc = new ArrayList();

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create the SQL statement to retrieve all readers
      StringBuffer sql =
          new StringBuffer("SELECT s.uid, u.username, s.last_read, s.last_post FROM confsettings s, "
			   + "users u WHERE u.uid = s.uid AND s.confid = ");
      sql.append(confid).append(" AND u.is_anon = 0 AND ISNULL(s.last_read) = 0 ORDER BY s.last_read DESC");
      if ((skip>=0) && (limit>0))
	sql.append(" LIMIT ").append(skip).append(", ").append(limit);
      sql.append(';');

      // execute the statement
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // return all the records as ActiveUser data elements
	ActiveUser usr = new ActiveUserImpl(rs.getInt(1),rs.getString(2),SQLUtil.getFullDateTime(rs,3),
					    SQLUtil.getFullDateTime(rs,4));
	rc.add(usr);

      } // end while

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error getting active reader list: " + e.getMessage(),e);
      throw new DataException("unable to get active reader listing: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return Collections.unmodifiableList(rc);

  } // end getActiveReaders

  public List getActiveReaders(int limit) throws DataException, AccessError
  {
    return getActiveReaders(0,limit);

  } // end getActiveReaders

  public List getActiveReaders() throws DataException, AccessError
  {
    return getActiveReaders(-1,-1);

  } // end getActiveReaders

  public List getMemberList() throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to get member list");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    return getConferenceData().getMemberList();

  } // end getMemberList

  public int getMemberLevel(int uid) throws DataException, AccessError
  {
    if (!(getConferenceData().canReadConference(level)))
    { // the luser can't even read the conference...
      logger.error("user not permitted to get member data");
      throw new AccessError("You are not permitted to read this conference.");

    } // end if

    return getConferenceData().getMemberLevel(uid);

  } // end getMemberLevel

  public void delete() throws DataException, AccessError
  {
    ConferenceSIGContext ctxt = getConferenceData();
    if (!(ctxt.canDeleteConference(level)) || !(sig.userCanDeleteSubobjects()))
    { // no way can we delete this conference, not NO way!
      logger.error("user not permitted to delete conference");
      throw new AccessError("You are not permitted to delete this conference.");

    } // end if

    // call the methods required to delete the conference
    ctxt.delete(sig);
    sig.detachConferenceDataObject(confid);

    // flag that we've been deleted
    cache = null;
    settings_loaded = true;
    pseud = null;
    last_read = null;
    last_post = null;
    deleted = true;

    // detach our references from the lower-level object
    confdata = null;
    ctxt.rd_release();

  } // end delete

  public boolean canDeleteConference()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return (c.canDeleteConference(level) && sig.userCanDeleteSubobjects());

  } // end canDeleteConference

  public boolean isInHotlist()
  {
    Connection conn = null;

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // do a quickie query
      StringBuffer sql = new StringBuffer("SELECT sequence FROM confhotlist WHERE uid = ");
      sql.append(sig.realUID()).append(" AND sigid = ").append(sig.realSIGID()).append(" AND confid = ");
      sql.append(confid).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      return rs.next();

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error checking hotlist: " + e.getMessage(),e);
      return false;

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end isInHotlist

  public void addToHotlist() throws DataException
  {
    Connection conn = null;

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      stmt.executeUpdate("LOCK TABLES confhotlist WRITE;");
      try
      { // do a quickie query to see if we're already in the hotlist
	StringBuffer sql = new StringBuffer("SELECT sequence FROM confhotlist WHERE uid = ");
	sql.append(sig.realUID()).append(" AND sigid = ").append(sig.realSIGID()).append(" AND confid = ");
	sql.append(confid).append(';');
	ResultSet rs = stmt.executeQuery(sql.toString());
	if (rs.next())
	  return;  // already in hotlist - this is a no-op

	// find a sequence number for the new entry
	sql.setLength(0);
	sql.append("SELECT MAX(sequence) FROM confhotlist WHERE uid = ").append(sig.realUID()).append(';');
	rs = stmt.executeQuery(sql.toString());
	if (!(rs.next()))
	  throw new InternalStateError("bogus query result on addToHotlist");
	int new_sequence = rs.getInt(1) + 100;

	// add the new record
	sql.setLength(0);
	sql.append("INSERT INTO confhotlist (uid, sequence, sigid, confid) VALUES (").append(sig.realUID());
	sql.append(", ").append(new_sequence).append(", ").append(sig.realSIGID()).append(", ");
	sql.append(confid).append(");");
	stmt.executeUpdate(sql.toString());

      } // end try
      finally
      { // make sure the table is unlocked before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error adding to hotlist: " + e.getMessage(),e);
      throw new DataException("error adding to hotlist: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end addToHotList

  public boolean canAddToHotlist()
  {
    if (sig.userIsAnonymous())
      return false;
    return !(isInHotlist());

  } // end canAddToHotlist

  public SIGContext getEnclosingSIG()
  {
    return sig.selfSIG();

  } // end getEnclosingSIG

  public void removeFromHotlist() throws DataException
  {
    Connection conn = null;

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create the DELETE statement and just execute it blind (if this conference is not in the hotlist,
      // the DELETE is a no-op).
      StringBuffer sql = new StringBuffer("DELETE FROM confhotlist WHERE uid = ");
      sql.append(sig.realUID()).append(" AND sigid = ").append(sig.realSIGID()).append(" AND confid = ");
      sql.append(confid).append(';');
      stmt.executeUpdate(sql.toString());

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error removing from hotlist: " + e.getMessage(),e);
      throw new DataException("error removing from hotlist: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end removeFromHotlist

  public void setHotlistSequence(int seq) throws DataException
  {
    Connection conn = null;

    try
    { // retrieve a connection from the datapool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create the UPDATE statement and just execute it blind (if this conference is not in the hotlist,
      // the UPDATE is a no-op).
      StringBuffer sql = new StringBuffer("UPDATE confhotlist SET sequence = ");
      sql.append(seq).append(" WHERE uid = ").append(sig.realUID()).append(" AND sigid = ");
      sql.append(sig.realSIGID()).append(" AND confid = ").append(confid).append(';');
      stmt.executeUpdate(sql.toString());

    } // end try
    catch (SQLException e)
    { // this becomes a DataException
      logger.error("DB error setting hotlist sequence: " + e.getMessage(),e);
      throw new DataException("error setting hotlist sequence: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setHotlistSequence

  /*--------------------------------------------------------------------------------
   * Implementations from interface UserBackend
   *--------------------------------------------------------------------------------
   */

  public int realUID()
  {
    return sig.realUID();

  } // end realUID

  public int realBaseLevel()
  {
    return sig.realBaseLevel();

  } // end realBaseLevel

  public String userRemoteAddress()
  {
    return sig.userRemoteAddress();

  } // end userRemoteAddress

  public String userDefaultPseud() throws DataException
  {
    return sig.userDefaultPseud();

  } // end userDefaultPseud

  public boolean userIsAnonymous()
  {
    return sig.userIsAnonymous();

  } // end userIsAnonymous

  public String realUserName()
  {
    return sig.realUserName();

  } // end realUserName

  public String realEmailAddress() throws DataException
  {
    return sig.realEmailAddress();

  } // end realEmailAddress

  public String realFullName() throws DataException
  {
    return sig.realFullName();

  } // end realFullName

  public void saveMRU(String tag, ReferencedData data)
  {
    sig.saveMRU(tag,data);

  } // end saveMRU

  /*--------------------------------------------------------------------------------
   * Implementations from interface SIGBackend
   *--------------------------------------------------------------------------------
   */

  public SIGContext selfSIG()
  {
    return sig.selfSIG();

  } // end selfsig;

  public int realSIGID()
  {
    return sig.realSIGID();

  } // end realSIGID

  public boolean userHideHiddenConferences()
  {
    return sig.userHideHiddenConferences();

  } // end userHideHiddenConferences

  public int realSIGLevel()
  {
    return sig.realSIGLevel();

  } // end realSIGLevel

  public ConferenceSIGContext getConferenceDataObject(int confid) throws DataException
  {
    return sig.getConferenceDataObject(confid);

  } // end getConferenceDataObject

  public boolean userCanCreateSubobjects()
  {
    return sig.userCanCreateSubobjects();

  } // end userCanCreateSubobjects

  public String realSIGAlias()
  {
    return sig.realSIGAlias();

  } // end realSIGAlias

  public void detachConferenceDataObject(int confid) throws DataException
  {
    sig.detachConferenceDataObject(confid);

  } // end detachConferenceDataObject

  public boolean userCanDeleteSubobjects()
  {
    return sig.userCanDeleteSubobjects();

  } // end userCanDeleteSubobjects

  /*--------------------------------------------------------------------------------
   * Implementations from interface ConferenceBackend
   *--------------------------------------------------------------------------------
   */

  public int realConfID()
  {
    return confid;

  } // end realConfID

  public boolean userCanHide()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canHideTopics(level);

  } // end userCanHide

  public void touchRead(Connection conn) throws SQLException
  {
    if (deleted || sig.userIsAnonymous())
      return;  // anonymous user can't update squat

    Statement stmt = conn.createStatement();
    StringBuffer sql = new StringBuffer();

    java.util.Date now = new java.util.Date();
    if (settings_loaded)
    { // generate an update statement
      sql.append("UPDATE confsettings SET last_read = '").append(SQLUtil.encodeDate(now));
      sql.append("' WHERE confid = ").append(confid).append(" AND uid = ").append(sig.realUID()).append(';');

    } // end if
    else
    { // need to insert a confsettings row
      sql.append("INSERT INTO confsettings (confid, uid, default_pseud, last_read) VALUES (").append(confid);
      sql.append(", ").append(sig.realUID()).append(", '").append(SQLUtil.encodeString(pseud)).append("', '");
      sql.append(SQLUtil.encodeDate(now)).append("');");

    } // end else

    // execute the statement
    stmt.executeUpdate(sql.toString());

    // save off the values to our local fields
    last_read = now;
    settings_loaded = true;

  } // end touchRead

  public void touchPost(Connection conn, java.util.Date post_date) throws SQLException
  {
    if (deleted || sig.userIsAnonymous())
      return;  // anonymous user can't update squat

    Statement stmt = conn.createStatement();
    StringBuffer sql = new StringBuffer();

    if (settings_loaded)
    { // generate an update statement
      sql.append("UPDATE confsettings SET last_post = '").append(SQLUtil.encodeDate(post_date));
      sql.append("' WHERE confid = ").append(confid).append(" AND uid = ").append(sig.realUID()).append(';');

    } // end if
    else
    { // need to insert a confsettings row
      sql.append("INSERT INTO confsettings (confid, uid, default_pseud, last_read) VALUES (").append(confid);
      sql.append(", ").append(sig.realUID()).append(", '").append(SQLUtil.encodeString(pseud)).append("', '");
      sql.append(SQLUtil.encodeDate(post_date)).append("');");

    } // end else

    // execute the statement
    stmt.executeUpdate(sql.toString());

    // save off the values to our local fields
    last_post = post_date;
    settings_loaded = true;

  } // end touchRead

  public String realConfAlias()
  {
    if (deleted)
      return null;

    try
    { // just pull it from the lower-level object
      return getConferenceData().getAnAlias();

    } // end try
    catch (DataException de)
    { // whoops - this wasn't supposed to happen!
      logger.error("DataException in realConfAlias(): " + de.getMessage(),de);
      return null;

    } // end catch

  } // end realConfAlias

  public boolean userCanScribble()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canScribblePosts(level);

  } // end userCanScribble

  public boolean userCanNuke()
  {
    ConferenceSIGContext c = getConferenceDataNE();
    if (c==null)
      return false;
    return c.canNukePosts(level);

  } // end userCanNuke

  public boolean userCanRead()
  {
    return canReadConference();

  } // end userCanRead

  public boolean userCanPost()
  {
    return canPostToConference();

  } // end userCanPost

  public void touchUpdate(Connection conn, java.util.Date date) throws DataException
  {
    getConferenceData().touchUpdate(conn,date);

  } // end getConferenceData

  public PostLinkDecoderContext createDecoderContext(short topicid)
  {
    return new PostLinkDecoderContext(sig.realSIGAlias(),realConfAlias(),topicid);

  } // end createDecoderContext

  /*--------------------------------------------------------------------------------
   * Static functions usable only from within the package
   *--------------------------------------------------------------------------------
   */

  static List getSIGConferences(EngineBackend engine, SIGBackend sig, DataPool datapool) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getSIGConferences for SIG # " + sig.realSIGID() + ", user #" + sig.realUID());
    ArrayList rc = new ArrayList();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // Build a monster query!  The idea is that we get the basic info of all conferences linked to the
      // SIG, as well as the user level (within the conference) that gets granted to all members of the
      // linked SIG.  (Commonly this is 0; the standard user level within the SIG will then take precedence.)
      // If the user has an entry within the confmember table for that conference, the level granted by
      // that entry is also reported.  The highest of these two levels and the level of the user within the
      // SIG is then taken to be the user's access level within the conference.
      StringBuffer sql =
	  new StringBuffer("SELECT c.confid, c.createdate, c.name, c.descr, s.granted_lvl, m.granted_lvl "
			   + "FROM sigtoconf s, confs c LEFT JOIN confmember m ON c.confid = m.confid "
			   + "AND m.uid = ");
      sql.append(sig.realUID()).append(" WHERE s.confid = c.confid AND s.sigid = ").append(sig.realSIGID());
      if (sig.userHideHiddenConferences())
	sql.append(" AND s.hide_list = 0");
      sql.append(" ORDER BY s.sequence, c.name;");
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // Run that monster query!
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // create the appropriate objects to return
	ConferenceContext cc = new ConferenceUserContextImpl(engine,sig,datapool,rs.getInt(1),rs.getString(3),
							     rs.getString(4),SQLUtil.getFullDateTime(rs,2),
							     rs.getInt(5),rs.getInt(6),conn);
	rc.add(cc);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading conference entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve conference information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return Collections.unmodifiableList(rc);

  } // end getSIGConferences

  static ConferenceContext getConference(EngineBackend engine, SIGBackend sig, DataPool datapool,
					 int confid) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getConference(#" + confid + ") for SIG # " + sig.realSIGID() + ", user #" + sig.realUID());

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // Build a monster query!  This is like the query in getSIGConferences, except that we add the
      // conference ID as a search term (thus we can return at most one row).
      StringBuffer sql =
	  new StringBuffer("SELECT c.createdate, c.name, c.descr, s.granted_lvl, m.granted_lvl "
			   + "FROM sigtoconf s, confs c LEFT JOIN confmember m ON c.confid = m.confid "
			   + "AND m.uid = ");
      sql.append(sig.realUID()).append(" WHERE s.confid = c.confid AND s.sigid = ");
      sql.append(sig.realSIGID()).append(" AND c.confid = ").append(confid).append(';');
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // Run that monster query!
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
	throw new DataException("conference ID#" + confid + " not found in SIG#" + sig.realSIGID());

      // pass back the new object
      return new ConferenceUserContextImpl(engine,sig,datapool,confid,rs.getString(2),
					   rs.getString(3),SQLUtil.getFullDateTime(rs,1),
					   rs.getInt(4),rs.getInt(5),conn);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading conference entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve conference information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getConference

  static ConferenceContext getConference(EngineBackend engine, SIGBackend sig, DataPool datapool,
					 String alias) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getConference(\"" + alias + "\") for SIG # " + sig.realSIGID() + ", user #"
		   + sig.realUID());

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // Build a monster query!  This is similar to the other getConference query, except that now we
      // have an extra table in the query (confalias) and we match on that.  Notice that pretty much
      // everything is joined on the conference ID...
      StringBuffer sql =
	  new StringBuffer("SELECT c.confid, c.createdate, c.name, c.descr, s.granted_lvl, m.granted_lvl "
			   + "FROM sigtoconf s, confalias a, confs c LEFT JOIN confmember m ON "
			   + "c.confid = m.confid AND m.uid = ");
      sql.append(sig.realUID()).append(" WHERE s.confid = c.confid AND a.confid = c.confid AND s.sigid = ");
      sql.append(sig.realSIGID()).append(" AND a.alias = '").append(SQLUtil.encodeString(alias));
      sql.append("';");
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // Run that monster query!
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
	throw new DataException("conference \"" + alias + "\" not found in SIG#" + sig.realSIGID());

      // pass back the new object
      return new ConferenceUserContextImpl(engine,sig,datapool,rs.getInt(1),rs.getString(3),rs.getString(4),
					   SQLUtil.getFullDateTime(rs,2),rs.getInt(5),rs.getInt(6),conn);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading conference entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve conference information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getConference

  static List getUserHotlist(EngineBackend engine, UserBackend user, DataPool datapool) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getUserHotlist for user #" + user.realUID());

    Connection conn = null;      // pooled database connection
    ArrayList rc = new ArrayList();    // return from this function

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // Build a monster query!  We pull the SIGID and ConfID entries from the confhotlist table,
      // but we need to pull in "confs" and "confmember" (left joined) to get enough data to create
      // a ConferenceUserContextImpl object.
      StringBuffer sql =
	new StringBuffer("SELECT h.sigid, h.confid, c.name, c.descr, c.createdate, s.granted_lvl, "
			 + "f.granted_lvl, h.sequence FROM confhotlist h, sigtoconf s, confs c "
			 + "LEFT JOIN confmember f ON h.confid = f.confid AND h.uid = f.uid "
			 + "WHERE h.confid = s.confid AND c.confid = h.confid AND h.uid = ");
      sql.append(user.realUID()).append(" ORDER BY h.sequence ASC;");
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // Run that puppy!
      ResultSet rs = stmt.executeQuery(sql.toString());

      // Create some temporary data structures to assist us in building the return list.
      HashMap sig_backend_cache = new HashMap();

      while (rs.next())
      { // retrieve the sigid from the resultset first
	Integer sigid = new Integer(rs.getInt(1));

	// we need a SIG backend for our conference, so make sure we have one
	SIGBackend sig = (SIGBackend)(sig_backend_cache.get(sigid));
	if (sig==null)
	{ // get it and make sure it's in the cache for next time
	  sig = SIGUserContextImpl.getSIGBackend(engine,user,datapool,conn,sigid.intValue());
	  sig_backend_cache.put(sigid,sig);

	} // end if

	// make the new ConferenceContext
	ConferenceContext conf =
	    new ConferenceUserContextImpl(engine,sig,datapool,rs.getInt(2),rs.getString(3),rs.getString(4),
					  SQLUtil.getFullDateTime(rs,5),rs.getInt(6),rs.getInt(7),conn);

	// and create the actual return value
	ConferenceHotlistEntry hle = new MyHotlist(conf,rs.getInt(8));
	rc.add(hle);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading hotlist entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve hotlist information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return Collections.unmodifiableList(rc);

  } // end getUserHotlist

} // end class ConferenceUserContextImpl
